import logging
from future.standard_library import install_aliases
install_aliases()
from urllib.request import urlopen

log = logging.getLogger(__name__)

SERVICE_NAME = 'ec2'
ORIGIN = 'AWS::EC2::Instance'


def initialize():
    """
    Try to get EC2 instance-id and AZ if running on EC2
    by querying http://169.254.169.254/latest/meta-data/.
    If not continue.
    """
    global runtime_context

    try:
        runtime_context = {}

        r = urlopen('http://169.254.169.254/latest/meta-data/instance-id', timeout=1)
        runtime_context['instance_id'] = r.read().decode('utf-8')

        r = urlopen('http://169.254.169.254/latest/meta-data/placement/availability-zone',
                    timeout=1)
        runtime_context['availability_zone'] = r.read().decode('utf-8')

    except Exception:
        runtime_context = None
        log.warning("failed to get ec2 instance metadata.")



import threading
from operator import attrgetter

TTL = 60 * 60  # The cache expires 1 hour after the last refresh time.


class RuleCache(object):
    """
    Cache sampling rules and quota retrieved by ``TargetPoller``
    and ``RulePoller``. It will not return anything if it expires.
    """
    def __init__(self):

        self._last_updated = None
        self._rules = []
        self._lock = threading.Lock()

    def get_matched_rule(self, sampling_req, now):
        if self._is_expired(now):
            return None
        matched_rule = None
        for rule in self.rules:
            if(not matched_rule and rule.match(sampling_req)):
                matched_rule = rule
            if(not matched_rule and rule.is_default()):
                matched_rule = rule
        return matched_rule

    def load_rules(self, rules):
        # Record the old rules for later merging.
        with self._lock:
            self._load_rules(rules)

    def load_targets(self, targets_dict):
        with self._lock:
            self._load_targets(targets_dict)

    def _load_rules(self, rules):
        oldRules = {}
        for rule in self.rules:
            oldRules[rule.name] = rule

        # Update the rules in the cache.
        self.rules = rules

        # Transfer state information to refreshed rules.
        for rule in self.rules:
            old = oldRules.get(rule.name, None)
            if old:
                rule.merge(old)

        # The cache should maintain the order of the rules based on
        # priority. If priority is the same we sort name by alphabet
        # as rule name is unique.
        self.rules.sort(key=attrgetter('priority', 'name'))

    def _load_targets(self, targets_dict):
        for rule in self.rules:
            target = targets_dict.get(rule.name, None)
            if target:
                rule.reservoir.load_quota(target['quota'],
                                          target['TTL'],
                                          target['interval'])
                rule.rate = target['rate']

    def _is_expired(self, now):
        # The cache is treated as expired if it is never loaded.
        if not self._last_updated:
            return True
        return now > self.last_updated + TTL

    @property
    def rules(self):
        return self._rules

    @rules.setter
    def rules(self, v):
        self._rules = v

    @property
    def last_updated(self):
        return self._last_updated

    @last_updated.setter
    def last_updated(self, v):
        self._last_updated = v



import copy
import logging
import importlib

from django.db import connections

from aws_xray_sdk.core import xray_recorder
from aws_xray_sdk.ext.dbapi2 import XRayTracedCursor

log = logging.getLogger(__name__)


def patch_db():
    for conn in connections.all():
        module = importlib.import_module(conn.__module__)
        _patch_conn(getattr(module, conn.__class__.__name__))


class DjangoXRayTracedCursor(XRayTracedCursor):
    def execute(self, query, *args, **kwargs):
        if xray_recorder.stream_sql:
            _previous_meta = copy.copy(self._xray_meta)
            self._xray_meta['sanitized_query'] = query
        result = super(DjangoXRayTracedCursor, self).execute(query, *args, **kwargs)
        if xray_recorder.stream_sql:
            self._xray_meta = _previous_meta
        return result

    def executemany(self, query, *args, **kwargs):
        if xray_recorder.stream_sql:
            _previous_meta = copy.copy(self._xray_meta)
            self._xray_meta['sanitized_query'] = query
        result = super(DjangoXRayTracedCursor, self).executemany(query, *args, **kwargs)
        if xray_recorder.stream_sql:
            self._xray_meta = _previous_meta
        return result

    def callproc(self, proc, args):
        if xray_recorder.stream_sql:
            _previous_meta = copy.copy(self._xray_meta)
            self._xray_meta['sanitized_query'] = proc
        result = super(DjangoXRayTracedCursor, self).callproc(proc, args)
        if xray_recorder.stream_sql:
            self._xray_meta = _previous_meta
        return result


def _patch_cursor(cursor_name, conn):
    attr = '_xray_original_{}'.format(cursor_name)

    if hasattr(conn, attr):
        log.debug('django built-in db {} already patched'.format(cursor_name))
        return

    if not hasattr(conn, cursor_name):
        log.debug('django built-in db does not have {}'.format(cursor_name))
        return

    setattr(conn, attr, getattr(conn, cursor_name))

    meta = {}

    if hasattr(conn, 'vendor'):
        meta['database_type'] = conn.vendor

    def cursor(self, *args, **kwargs):

        host = None
        user = None

        if hasattr(self, 'settings_dict'):
            settings = self.settings_dict
            host = settings.get('HOST', None)
            user = settings.get('USER', None)

        if host:
            meta['name'] = host
        if user:
            meta['user'] = user

        original_cursor = getattr(self, attr)(*args, **kwargs)
        return DjangoXRayTracedCursor(original_cursor, meta)

    setattr(conn, cursor_name, cursor)


def _patch_conn(conn):
    _patch_cursor('cursor', conn)
    _patch_cursor('chunked_cursor', conn)



import wrapt
import botocore.client

from aws_xray_sdk.core import xray_recorder
from aws_xray_sdk.ext.boto_utils import inject_header, aws_meta_processor


def patch():
    """
    Patch botocore client so it generates subsegments
    when calling AWS services.
    """
    if hasattr(botocore.client, '_xray_enabled'):
        return
    setattr(botocore.client, '_xray_enabled', True)

    wrapt.wrap_function_wrapper(
        'botocore.client',
        'BaseClient._make_api_call',
        _xray_traced_botocore,
    )

    wrapt.wrap_function_wrapper(
        'botocore.endpoint',
        'Endpoint.prepare_request',
        inject_header,
    )


def _xray_traced_botocore(wrapped, instance, args, kwargs):
    service = instance._service_model.metadata["endpointPrefix"]
    if service == 'xray':
        # skip tracing for SDK built-in sampling pollers
        if 'GetSamplingRules' in args or 'GetSamplingTargets' in args:
            return wrapped(*args, **kwargs)
    return xray_recorder.record_subsegment(
        wrapped, instance, args, kwargs,
        name=service,
        namespace='aws',
        meta_processor=aws_meta_processor,
    )



import inspect
import sys


PY2 = sys.version_info < (3,)
PY35 = sys.version_info >= (3, 5)

if PY2:
    annotation_value_types = (int, long, float, bool, str)  # noqa: F821
    string_types = basestring  # noqa: F821
else:
    annotation_value_types = (int, float, bool, str)
    string_types = str


def is_classmethod(func):
    return getattr(func, '__self__', None) is not None


def is_instance_method(parent_class, func_name, func):
    try:
        func_from_dict = parent_class.__dict__[func_name]
    except KeyError:
        for base in inspect.getmro(parent_class):
            if func_name in base.__dict__:
                func_from_dict = base.__dict__[func_name]
                break
        else:
            return True

    return not is_classmethod(func) and not isinstance(func_from_dict, staticmethod)



import logging
import socket

from aws_xray_sdk.core.daemon_config import DaemonConfig
from ..exceptions.exceptions import InvalidDaemonAddressException

log = logging.getLogger(__name__)


PROTOCOL_HEADER = "{\"format\":\"json\",\"version\":1}"
PROTOCOL_DELIMITER = '\n'
DEFAULT_DAEMON_ADDRESS = '127.0.0.1:2000'


class UDPEmitter(object):
    """
    The default emitter the X-Ray recorder uses to send segments/subsegments
    to the X-Ray daemon over UDP using a non-blocking socket. If there is an
    exception on the actual data transfer between the socket and the daemon,
    it logs the exception and continue.
    """
    def __init__(self, daemon_address=DEFAULT_DAEMON_ADDRESS):

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.setblocking(0)
        self.set_daemon_address(daemon_address)

    def send_entity(self, entity):
        """
        Serializes a segment/subsegment and sends it to the X-Ray daemon
        over UDP. By default it doesn't retry on failures.

        :param entity: a trace entity to send to the X-Ray daemon
        """
        message = "%s%s%s" % (PROTOCOL_HEADER,
                              PROTOCOL_DELIMITER,
                              entity.serialize())

        log.debug("sending: %s to %s:%s." % (message, self._ip, self._port))
        self._send_data(message)

    def set_daemon_address(self, address):
        """
        Set up UDP ip and port from the raw daemon address
        string using ``DaemonConfig`` class utlities.
        """
        if address:
            daemon_config = DaemonConfig(address)
            self._ip, self._port = daemon_config.udp_ip, daemon_config.udp_port

    @property
    def ip(self):
        return self._ip

    @property
    def port(self):
        return self._port

    def _send_data(self, data):

        try:
            self._socket.sendto(data.encode('utf-8'), (self._ip,
                                self._port))
        except Exception:
            log.exception('failed to send data to X-Ray daemon.')

    def _parse_address(self, daemon_address):
        try:
            val = daemon_address.split(':')
            return val[0], int(val[1])
        except Exception:
            raise InvalidDaemonAddressException('Invalid daemon address %s specified.' % daemon_address)



import logging

from django.template import Template

from aws_xray_sdk.core import xray_recorder

log = logging.getLogger(__name__)


def patch_template():

    attr = '_xray_original_render'

    if getattr(Template, attr, None):
        log.debug("already patched")
        return

    setattr(Template, attr, Template.render)

    @xray_recorder.capture('template_render')
    def xray_render(self, context):
        template_name = self.name or getattr(context, 'template_name', None)
        if template_name:
            name = str(template_name)
            subsegment = xray_recorder.current_subsegment()
            if subsegment:
                subsegment.name = name

        return Template._xray_original_render(self, context)

    Template.render = xray_render



import logging
from random import Random
import time
import threading

from .local.sampler import LocalSampler
from .rule_cache import RuleCache
from .rule_poller import RulePoller
from .target_poller import TargetPoller
from .connector import ServiceConnector
from .reservoir import ReservoirDecision
from aws_xray_sdk import global_sdk_config

log = logging.getLogger(__name__)


class DefaultSampler(object):
    """Making sampling decisions based on centralized sampling rules defined
    by X-Ray control plane APIs. It will fall back to local sampler if
    centralized sampling rules are not available.
    """
    def __init__(self):
        self._local_sampler = LocalSampler()
        self._cache = RuleCache()
        self._connector = ServiceConnector()
        self._rule_poller = RulePoller(self._cache, self._connector)
        self._target_poller = TargetPoller(self._cache,
                                           self._rule_poller, self._connector)

        self._xray_client = None
        self._random = Random()
        self._started = False
        self._origin = None
        self._lock = threading.Lock()

    def start(self):
        """
        Start rule poller and target poller once X-Ray daemon address
        and context manager is in place.
        """
        if not global_sdk_config.sdk_enabled():
            return

        with self._lock:
            if not self._started:
                self._rule_poller.start()
                self._target_poller.start()
                self._started = True

    def should_trace(self, sampling_req=None):
        """
        Return the matched sampling rule name if the sampler finds one
        and decide to sample. If no sampling rule matched, it falls back
        to the local sampler's ``should_trace`` implementation.
        All optional arguments are extracted from incoming requests by
        X-Ray middleware to perform path based sampling.
        """
        if not global_sdk_config.sdk_enabled():
            return False

        if not self._started:
            self.start() # only front-end that actually uses the sampler spawns poller threads

        now = int(time.time())
        if sampling_req and not sampling_req.get('service_type', None):
            sampling_req['service_type'] = self._origin
        elif sampling_req is None:
            sampling_req = {'service_type': self._origin}
        matched_rule = self._cache.get_matched_rule(sampling_req, now)
        if matched_rule:
            log.debug('Rule %s is selected to make a sampling decision.', matched_rule.name)
            return self._process_matched_rule(matched_rule, now)
        else:
            log.info('No effective centralized sampling rule match. Fallback to local rules.')
            return self._local_sampler.should_trace(sampling_req)

    def load_local_rules(self, rules):
        """
        Load specified local rules to local fallback sampler.
        """
        self._local_sampler.load_local_rules(rules)

    def load_settings(self, daemon_config, context, origin=None):
        """
        The pollers have dependency on the context manager
        of the X-Ray recorder. They will respect the customer
        specified xray client to poll sampling rules/targets.
        Otherwise they falls back to use the same X-Ray daemon
        as the emitter.
        """
        self._connector.setup_xray_client(ip=daemon_config.tcp_ip,
                                          port=daemon_config.tcp_port,
                                          client=self.xray_client)

        self._connector.context = context
        self._origin = origin

    def _process_matched_rule(self, rule, now):
        # As long as a rule is matched we increment request counter.
        rule.increment_request_count()
        reservoir = rule.reservoir
        sample = True
        # We check if we can borrow or take from reservoir first.
        decision = reservoir.borrow_or_take(now, rule.can_borrow)
        if(decision == ReservoirDecision.BORROW):
            rule.increment_borrow_count()
        elif (decision == ReservoirDecision.TAKE):
            rule.increment_sampled_count()
        # Otherwise we compute based on fixed rate of this sampling rule.
        elif (self._random.random() <= rule.rate):
            rule.increment_sampled_count()
        else:
            sample = False

        if sample:
            return rule.name
        else:
            return False

    @property
    def xray_client(self):
        return self._xray_client

    @xray_client.setter
    def xray_client(self, v):
        self._xray_client = v



"""
AioHttp Client tracing, only compatible with Aiohttp 3.X versions
"""
import aiohttp

from types import SimpleNamespace

from aws_xray_sdk.core import xray_recorder
from aws_xray_sdk.core.models import http
from aws_xray_sdk.core.utils import stacktrace
from aws_xray_sdk.ext.util import inject_trace_header, strip_url

# All aiohttp calls will entail outgoing HTTP requests, only in some ad-hoc
# exceptions the namespace will be flip back to local.
REMOTE_NAMESPACE = 'remote'
LOCAL_NAMESPACE = 'local'
LOCAL_EXCEPTIONS = (
    aiohttp.client_exceptions.ClientConnectionError,
    # DNS issues
    OSError
)


async def begin_subsegment(session, trace_config_ctx, params):
    name = trace_config_ctx.name if trace_config_ctx.name else strip_url(str(params.url))
    subsegment = xray_recorder.begin_subsegment(name, REMOTE_NAMESPACE)

    # No-op if subsegment is `None` due to `LOG_ERROR`.
    if not subsegment:
        trace_config_ctx.give_up = True
    else:
        trace_config_ctx.give_up = False
        subsegment.put_http_meta(http.METHOD, params.method)
        subsegment.put_http_meta(http.URL, params.url.human_repr())
        inject_trace_header(params.headers, subsegment)


async def end_subsegment(session, trace_config_ctx, params):
    if trace_config_ctx.give_up:
        return

    subsegment = xray_recorder.current_subsegment()
    subsegment.put_http_meta(http.STATUS, params.response.status)
    xray_recorder.end_subsegment()


async def end_subsegment_with_exception(session, trace_config_ctx, params):
    if trace_config_ctx.give_up:
        return

    subsegment = xray_recorder.current_subsegment()
    subsegment.add_exception(
        params.exception,
        stacktrace.get_stacktrace(limit=xray_recorder._max_trace_back)
    )

    if isinstance(params.exception, LOCAL_EXCEPTIONS):
        subsegment.namespace = LOCAL_NAMESPACE

    xray_recorder.end_subsegment()


def aws_xray_trace_config(name=None):
    """
    :param name: name used to identify the subsegment, with None internally the URL will
                 be used as identifier.
    :returns: TraceConfig.
    """

    def _trace_config_ctx_factory(trace_request_ctx):
        return SimpleNamespace(
            name=name,
            trace_request_ctx=trace_request_ctx
        )

    trace_config = aiohttp.TraceConfig(trace_config_ctx_factory=_trace_config_ctx_factory)
    trace_config.on_request_start.append(begin_subsegment)
    trace_config.on_request_end.append(end_subsegment)
    trace_config.on_request_exception.append(end_subsegment_with_exception)
    return trace_config



import copy
import re
import wrapt
from operator import methodcaller

from aws_xray_sdk.ext.dbapi2 import XRayTracedConn, XRayTracedCursor


def patch():

    wrapt.wrap_function_wrapper(
        'psycopg2',
        'connect',
        _xray_traced_connect
    )
    wrapt.wrap_function_wrapper(
        'psycopg2.extensions',
        'register_type',
        _xray_register_type_fix
    )


def _xray_traced_connect(wrapped, instance, args, kwargs):

    conn = wrapped(*args, **kwargs)
    parameterized_dsn = { c[0]: c[-1] for c in map(methodcaller('split', '='), conn.dsn.split(' '))}
    meta = {
        'database_type': 'PostgreSQL',
        'url': 'postgresql://{}@{}:{}/{}'.format(
            parameterized_dsn.get('user', 'unknown'),
            parameterized_dsn.get('host', 'unknown'),
            parameterized_dsn.get('port', 'unknown'),
            parameterized_dsn.get('dbname', 'unknown'),
        ),
        'user': parameterized_dsn.get('user', 'unknown'),
        'database_version': str(conn.server_version),
        'driver_version': 'Psycopg 2'
    }

    return XRayTracedConn(conn, meta)

def _xray_register_type_fix(wrapped, instance, args, kwargs):
    """Send the actual connection or curser to register type."""
    our_args = list(copy.copy(args))
    if len(our_args) == 2 and isinstance(our_args[1], (XRayTracedConn, XRayTracedCursor)):
        our_args[1] = our_args[1].__wrapped__

    return wrapped(*our_args, **kwargs)



from builtins import super
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.session import Session, sessionmaker
from .util.decorators import xray_on_call, decorate_all_functions


@decorate_all_functions(xray_on_call)
class XRaySession(Session):
    pass


@decorate_all_functions(xray_on_call)
class XRayQuery(Query):
    pass


@decorate_all_functions(xray_on_call)
class XRaySessionMaker(sessionmaker):
    def __init__(self, bind=None, class_=XRaySession, autoflush=True,
                 autocommit=False,
                 expire_on_commit=True,
                 info=None, **kw):
        kw['query_cls'] = XRayQuery
        super().__init__(bind, class_, autoflush, autocommit, expire_on_commit,
                         info, **kw)
